#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>
#include <selinux/selinux.h> 	/* for getpidcon function */

#include "jobd.h"



struct security_context_list{
	
		char security_context[buffer_size];	// security context assosicated
		int elsa_group[MAX_GROUPS_PER_RULE];	// list of jobs with which given security context is assosiated.
		struct security_context_list *security_context_nxt;	// link list pointer for next.

};

struct security_context_list *head;	// global head pointer which points to start of linked list of security context rules.



void parse_group(char *string,int groups[]);




/* file_parse function to parse the config file that has the information 
	about security context and corresponding elsa group 
*/
struct security_context_list * file_parse () 
{
	FILE *fd;
	char buff[buffer_size+1];
	char *parse_ptr, *save_ptr,*tmp,*tmp_str;
	char delimiter[] = " ";
	struct security_context_list *security_context_tmp,*security_context_tmp1,*security_context_tmp2;
	int len;

	head =NULL;
	
	//To open the config the file
	fd = fopen(CONFIG_FILE_PATH, "r");
	if (fd == NULL)
	{
		printf("Error: not able to open configuration file %s in read mode\nCheck if file is present\n", CONFIG_FILE_PATH );
		exit (1);
	}
	
	while(fgets(buff,buffer_size,fd) != NULL)
	{
		len = strlen (buff);
		if (len > 0 )
		{
			if ( buff[len-1]== '\n')
		        {
		        	buff[len-1] = '\0';
		        }
		}
		else
		{
		 	 continue ;
		}
		parse_ptr = buff;

		  // check if string is empty or not
		  for (; *parse_ptr ==' ' || *parse_ptr == '\t' ||  *parse_ptr == '\v'  ; ++parse_ptr );
		  if (parse_ptr == NULL )
		  {
		  	if( feof (fd))
			{
				break ;
			}
			continue ;
		}
		parse_ptr = buff;
		if(*parse_ptr == '#')
		{
			continue;
		}

		security_context_tmp = malloc(sizeof(struct security_context_list));
		parse_ptr = buff;

		tmp = strtok_r(parse_ptr,delimiter,&save_ptr);
		strcpy (security_context_tmp->security_context,tmp);
		security_context_tmp->security_context_nxt = NULL;
		
		if(head == NULL)
		{
			head = security_context_tmp;
		}
		else
		{
			security_context_tmp1 = head;
			while(security_context_tmp1 != NULL)
			{
				security_context_tmp2 = security_context_tmp1;
				security_context_tmp1 = security_context_tmp1->security_context_nxt;
			}
			security_context_tmp1 = security_context_tmp;
			security_context_tmp2->security_context_nxt = security_context_tmp1;
		}
		
		parse_ptr = NULL;
		tmp = NULL;
		
		
		tmp = strtok_r(parse_ptr,delimiter,&save_ptr);
		tmp_str=malloc(sizeof(strlen(tmp)));
		strcpy(tmp_str,tmp);

		 
		parse_group(tmp_str,security_context_tmp->elsa_group);


	} //end of while
	
	return head;
	

}

//To assign values of elsa groups in the array of security_context_list 
// We assume that one process can belong to more than one jobs.
// this function will parse comma seperated list of groupids.
void parse_group(char *string,int groups[])

{
	char *tmp_pt,*save_ptr,*tmp_ptr;
	char delimiter[] = ",";
	int i,group_id;
	
	i = 0 ;
	for(tmp_ptr=string; ; tmp_ptr=NULL)
	{
		tmp_pt=strtok_r(tmp_ptr,delimiter,&save_ptr);
		if(tmp_pt == NULL)
		{
			break;
		}
		group_id = atoi(tmp_pt);
		groups[i] = group_id;	
		i++;
		if ( i >= MAX_GROUPS_PER_RULE )
			break ;
		tmp_ptr = NULL ;
	}
}


/*
Finds the largest job number specified in configuration.
This information is required so that, all those empty groups can be created at system initialization
*/

int find_default_group_number()
{
	int i,max;
	struct security_context_list *security_context_tmp;

	security_context_tmp = head;
	max=0;

	while(security_context_tmp != NULL)
	{
		for(i=0; (i < MAX_GROUPS_PER_RULE) && ( security_context_tmp->elsa_group[i] > 0 ) ;i++)
		{
			if(security_context_tmp->elsa_group[i] > max)
			{
				max = security_context_tmp->elsa_group[i];
			}
		}
		security_context_tmp = security_context_tmp->security_context_nxt;
	}
	return max;
}



////////////////////////////////////////////////////

//search_security_context function searches given s_context
//in provided policies with considering '*' as wild-card.
//This wildcard allows you to skip some part of policy matching.
//You are allowed to specify policies in following manner.

//user_t:role_t:type_t       5
//user_t:*:type_t	     7
// *:role_t:*	             2,4
// *:*:*		     1,3

// Order of rules is important, put specific rules first, and then put generic rules
// on success returns the pointer to security_context_list structure containing matching rule
// on failure returns NULL

struct security_context_list *  search_security_context ( char * s_context )
{
        int is_failed ;
        char *rule_part, *context_part ;
	struct security_context_list *temp1 ;

        if ( s_context == NULL || !(*s_context) ) return NULL ;



	for (temp1=head; temp1 !=NULL ; temp1=temp1->security_context_nxt)
        {
                rule_part = temp1->security_context ;
                context_part = s_context ;
                is_failed = 0 ;

                // if this rule is empty, following rules also should be empty then.
                if (!(*rule_part)) break ; // it means, all rule after this one are empty.
//                JOBD_DEBUG (  "Rule no %d present\n",rule_number );
//                JOBD_DEBUG (  "Comparing %s = %s \n", rule_part, s_context);


                // seperate and compare the user.
                if ( *rule_part == '*' )
                {
                        // if it is '*', then its wild character, and dont need to be matched
                        // forword the rule part till next ':'
                        for ( ; *rule_part && *rule_part != ':' ; rule_part++ ) ;

                        // If there is unexpected end of line then throw error
                        if ( !(*rule_part) )
                        {
                        	JOBD_DEBUG  ( "Error in rule of rule base %s \n",
                                       temp1->security_context) ;
                                is_failed = 1 ;
                                continue ;
                        }

                        // forword the context part till next ':'
                        for ( ; *context_part && *context_part != ':' ; context_part++ ) ;

                        // If there is unexpected end of line then throw error
                        if ( !(*context_part) )
                        {
                              JOBD_DEBUG ( "Error in matching rule of rule base %s with context %s\n",
                                        temp1->security_context, s_context) ;
                                is_failed = 1 ;
                                continue ;
                        }
                }
                else
                {
                        // Its not wild-char now. so compare rule_part and context_part character by character.
                        // till next ':'
                        for ( ; *rule_part && *rule_part != ':' ; rule_part++, context_part++ )
                        {
                                if ( *rule_part != *context_part )
                                {
                                        is_failed = 1 ;
                                        break ;
                                }
                        } // end for: till next ':'
                        if (is_failed) continue ;

                        // If any of string terminates unexpectedly, then return with error
                        if ( !(*rule_part && *context_part))
                        {
                                is_failed = 1 ;
                                continue ;
                        }
                } // end if: if '*'

                // move and pass the seperator ':'
                rule_part++ ;
                context_part++ ;


                // seperate and compare the role.
                if ( *rule_part == '*' )
                {
                        // if it is '*', then its wild character, and dont need to be matched
                        // forword the rule part till next ':'
                        for ( ; *rule_part && *rule_part != ':' ; rule_part++ ) ;

                        // If there is unexpected end of line then throw error
                        if ( !(*rule_part) )
                        {
                              JOBD_DEBUG ( "Error in rule of rule base %s \n",temp1->security_context) ;
                                is_failed = 1 ;
                                continue ;
                        }

                        // forword the context part till next ':'
                        for ( ; *context_part && *context_part != ':' ; context_part++ ) ;

                        // If there is unexpected end of line then throw error
                        if ( !(*context_part) )
                        {
                              JOBD_DEBUG ( "Error in matching rule of rule base %s with context %s\n", temp1->security_context, s_context ) ;
                                is_failed = 1 ;
                                continue ;
                        }
                }
                else
                {
                        // Its not wild-char now. so compare rule_part and context_part character by character.
                        // till next ':'
                        for ( ; *rule_part && *rule_part != ':' ; rule_part++,context_part++ )
                        {
                                if ( *rule_part != *context_part )
                                {
                                        is_failed = 1 ;
                                        break ;
                                }
                        } // end for: till next ':'
                        // If any of string terminates unexpectedly, then return with error

                        if ( is_failed ) continue ;

                        if ( !(*rule_part && *context_part))
                        {
                                is_failed = 1 ;
                                continue ;
                        }
                }
                // move and pass the seperator ':'
                rule_part++ ;
                context_part++ ;

                // seperate and compare the domain.
                if ( *rule_part == '*' )
                {
                        // wildcard match successfull :-)
                        return  temp1 ;
                }
                else
                {
                        // Its not wild-char now. so compare rule_part and context_part character by character.
                        // till next ':'
                        for ( ; *rule_part && *rule_part != ':' ; rule_part++,context_part++ )
                        {
                                if ( *rule_part != *context_part )
                                {
                                        is_failed = 1 ;
                                        break ;
                                }
                        } // end for: till next ':'

                        if (is_failed ) continue ;

                        // match successfull :-)
                        return  temp1 ;
                }
        } // end for : for each rule in rule base.
        return NULL;
}

/*
Function classify_process will take the pid and will return the list of group to which this process should belong.
on success, it returns array of job_id's
on failure, it returns NULL.
*/
int *  get_process_classification ( int pid )
{
        struct security_context_list * matched_rule ;
	int ret, i ;
        security_context_t s_context ;


        ret = getpidcon ( pid, &s_context ) ;
        if ( ret == -1 )
        {
		JOBD_DEBUG("[get_process_classification] getpidcon failed to get s_context for %d", pid );
		return NULL ;
        }
	// now ret will contain jobid in which process should go.
	matched_rule = search_security_context (s_context) ;
	if (matched_rule == NULL )
	{
		return NULL ;
	}

	for ( i = 0 ;(i< MAX_GROUPS_PER_RULE)&& ( matched_rule->elsa_group[i] > 0 ) ; ++i )
	{
		JOBD_DEBUG("[get_process_classification] [%d => %s]==>J# %d", pid, s_context, matched_rule->elsa_group[i] );
	}		
	return matched_rule->elsa_group ;
} // end function :
