/*
 * bsdacct_ver.c
 *
 * This program tests the BSD Process Accounting version that is
 * used on the machine if any. It returns the version or -1 if an
 * error occured. 
 *
 * Copyright (C) 2006 BULL SA.
 * Written by Guillaume Thouvenin <guillaume.thouvenin@bull.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>

#include <sys/acct.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>

#define acctfile "/tmp/bsdacct.tmp"

int main()
{
	int acct_fd;
	int retval = 0;
	int status;
	pid_t pid;
	FILE *acct_fp;
	char buffer[2];

	/* You must be root to start accounting */
	if (getuid() != 0) {
		printf("You must be root to start accounting\n");
		return -1;
	}

	/* Start process accounting */
	/* We create a file to record accounting data */
	if ((acct_fd = creat(acctfile, S_IRUSR | S_IWUSR)) == -1) {
		perror("Cannot create accouting file");
		return acct_fd;
	}
	close(acct_fd);

	if ((retval = acct(acctfile)) == -1) {
		perror("Cannot start process accounting");
		goto fini;
	}

	/* To test the version of the BSD accounting we create a child 
	 * process and we check the ac_version field in the accounting 
	 * file. The goal is to have a non-empty file.
	 */
	pid = fork();
	switch (pid) {
	case 0:
		/* Child */
		return 0;
	case -1:
		/* Error */
		perror("Cannot fork");
		retval = -1;
		goto fini;
	default:
		/* parent */
		waitpid(pid, &status, 0);
	}

	/* Stop process accouting */
	acct(NULL);

	/* Now we can read the accounting file and look for the version */
	if ((acct_fp = fopen(acctfile, "r")) == NULL) {
		perror("Cannot read accounting data");
		return -1;
	}

	if (fread(buffer, sizeof(char), sizeof(buffer), acct_fp) !=
	    sizeof(buffer)) {
		perror("Cannot read the first two items");
		fclose(acct_fp);
		return -1;
	} else {
		printf(">>> BSD Process Accounting version %d <<< \n",
		       buffer[1]);
		retval = (int)buffer[1];
	}

	/* Close the stream and delete the file */
	fclose(acct_fp);
fini:
	unlink(acctfile);
	return retval;
}
